/*
 * =====================================================================================
 *
 *       Filename:  ConvertTed2DfgMinLatency.cc
 *    Description:
 *        Created:  09/10/2009 01:18:50 PM
 *         Author:  Daniel f. Gomez-Prado (dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 251                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2012-08-18 21:26:36 -0400 (Sat, 18 Aug 2012)     $: Date of last commit
 * =====================================================================================
 */

#include <cassert>
#include <string>
#include <iostream>
#include <bitset>
using namespace std;

#include "TedMan.h"
#include "TedNode.h"
#include "TedParents.h"
using namespace ted;

#include "DfgMan.h"
#include "DfgNode.h"
using namespace dfg;

#include "ConvertTed2DfgMinLatency.h"

namespace convert {

	ConvertTed2DfgMinLatency::ConvertTed2DfgMinLatency(TedMan * tm, DfgMan* dm) : \
		_tMan(tm), _dMan(dm), _dOne(NULL), _piMapR(dm), _dMapR(dm), _dMapP(dm), _tMap(dm) {
			_dOne = _tMap.get(1);
			_tMan->preOrderDFS(_parents, &TedParents::collect);
			_respect_ntl_level = false;
		}

	void ConvertTed2DfgMinLatency::respect_ntl_level(bool condition) {
		_respect_ntl_level = condition;
	}
	/*
	   unsigned int ConvertTed2DfgMinLatency::latency(void) const {
	   unsigned int total_latency = 0;
	   if (!paths.empty())
	   total_latency = paths.rbegin()->first;
	   return total_latency;
	   }*/

	void ConvertTed2DfgMinLatency::translate(void) {
		//
		// Now we are doing a DFS traversal, so that all nodes traversed have their bottom
		// DFG counterpart already created, this should speed things up.
		//
		// register outputs are assumed to be available as primary inputs therefore
		// its delay is disregarded for the computation of the critical latency.
		// tds::Environment::getInt(tds::Environment::DELAYREG);
		//
		for (TedMan::dfs_iterator it= _tMan->dfs_begin(); it != _tMan->dfs_end(); it++) {
			TedNode* node = (*it);
			if (node==TedNode::getOne()) {
				continue;
			}
			translate(node);
		}

		//
		// now that everything has been translated we look for the outputs and the weight
		// and registers associated to each output.
		//
		for(TedMan::PrimaryOutputs::iterator it = _tMan->_pos.begin(); it!=_tMan->_pos.end(); it++ ) {
			TedNode* tNode = it->second.Node();
			DfgNode* dNode = _tMap.get(tNode,1);
			int weight = it->second.getWeight();
			DfgNode* dout = NULL;
			if (weight!= 1) {
				DfgNode* constant = new DfgNode(_dMan,weight);
				dout = dNode->mul(constant);
			} else {
				dout = dNode;
			}
			TedRegister retimed = it->second.getRegister();
			if (retimed>0) {
				if (!_dMapR.has(dout)) {
					_dMapR.registrate(dout,1,dout->reg(1));
				}
				dout = _dMapR.get(dout,retimed);
			}
			_dMan->_mPos[it->first] = dout;
		}
	}

	DfgNode* ConvertTed2DfgMinLatency::getPIR(TedNode* pNode,int registers) {
		string piName = pNode->getVar()->getRootName();
		if (!_piMapR.has(piName)) {
			DfgNode* currentNode = NULL;
			switch (pNode->getVar()->getType()) {
			case TedVar::CONST:
				{
					currentNode = new DfgNode(_dMan,piName,DfgOperator::CONST,pNode->getVar()->getValue());
					break;
				}
			case TedVar::VARCONST:
				{
					currentNode = new DfgNode(_dMan,piName,DfgOperator::VARCONST,pNode->getVar()->getValue());
					break;
				}
			case TedVar::VAR:
				{
					currentNode = new DfgNode(_dMan,piName,DfgOperator::VAR);
					break;
				}
			case TedVar::VARBINARY:
				{
					currentNode = new DfgNode(_dMan,piName,DfgOperator::VAR);
					break;
				}
			case TedVar::ONEHOT:
			default:
				assert(false);
			}
			_piMapR.registrate(piName,0,currentNode);
		}
		return _piMapR.get(piName,registers);
	}

	DfgNode* ConvertTed2DfgMinLatency::resolve_mul_tree(DfgChain& muls2, wedge w, bool prevent_invertion) {
		assert(muls2.size() ==2);
		DfgChain::iterator ms1 = muls2.begin();
		DfgChain::iterator ms2 = ms1;
		ms2++;
		bool invert = false;
		if (!prevent_invertion) {
			invert = (w<0);
			w *= (invert) ? -1 : 1;
		}
		DfgNode* result = NULL;
		if (w!=1) {
			result = ms1->second->mul(_tMap.get(w));
		} else {
			result = ms1->second;
		}
		result = result->mul(ms2->second);
		result = reinterpret_cast<DfgNode*>( (long)result | invert);
		return result;
	}

	pair<bool, bool> ConvertTed2DfgMinLatency::belong_to_chain(TedNode* node) {
		//
		// Find if the node belongs to a chained operation
		//
		bool connects_to_additive_tree = false;
		bool connects_to_multiplicative_tree = false;
		if ((_parents.numParents(node) ==1) && !_tMan->isPO(node) && node != TedNode::getOne()) {
			MyParent::iterator parent_it = _parents[node]->begin();
			unsigned int parent_index = parent_it->second->at(0);
			assert(parent_it->first->hasKid(parent_index));
			if (parent_it->first->getKidRegister(parent_index) ==0) {
				if (parent_index==0) {
					//additive could have any number of kids but parent must have unitary weight
					connects_to_additive_tree = (parent_it->first->getKidWeight(0) ==1 || parent_it->first->getKidWeight(0) ==-1);
				} else if (node->numKids() ==1) {
					//multiplicative could have any parent index other than 0, but must have only one kid
					connects_to_multiplicative_tree = true;
				}
			}
		}
		return pair<bool,bool>(connects_to_multiplicative_tree,connects_to_additive_tree);
	}

	void ConvertTed2DfgMinLatency::translate(TedNode* node) {
		//
		// Find the highest retime node with the same parent variable
		//
		TedRegister node_reg = node->getRegister();
		DfgNode* current_dnode = getPIR(node,node_reg);

		pair<bool,bool> tree_op = belong_to_chain(node);
		bool connect_to_multiplicative_tree = tree_op.first;
		bool connect_to_additive_tree = tree_op.second;
		//
		// multiplicative chains is the simplest case
		if (connect_to_multiplicative_tree) {
			assert(!_tMan->isPO(node));
			unsigned int index = node->getKids()->highestOrder();
			assert(index!=0);
			current_dnode = _dMapP.get(current_dnode,index);
			unsigned int current_dnode_delay = ceil(log10((float)(index))/log10((float)2))*DfgNode::delayMul();
			current_dnode_delay += (_respect_ntl_level) ? node->getVar()->getLevel() : 0;

			DfgChain to_be_multiplied;
			to_be_multiplied.insert(DfgChain::value_type(current_dnode_delay,current_dnode));
			TedNode* kid_node = node->getKidNode(index);
			wedge weight = node->getKidWeight(index);
			assert(weight==1);
			TedRegister reg = node->getKidRegister(index);

			if (delayMultiplicative.find(kid_node)!=delayMultiplicative.end()) {
				// kid already belongs to a multiplicative tree
				DfgChain& mul_chain = delayMultiplicative[kid_node];
				to_be_multiplied.insert(mul_chain.begin(),mul_chain.end());
			} else {
				// this is the beginning of a multiplicative tree
				if (kid_node != TedNode::getOne()) {
					DfgNode* kid = _tMap.get(kid_node,1);
					unsigned int kid_delay = delayMap[kid_node];
					if (reg>0) {
						kid = _dMapR.get(kid,reg);
						kid_delay = 0;
					}
					to_be_multiplied.insert(DfgChain::value_type(kid_delay,kid));
				}
			}
			delayMultiplicative.insert(MapChain::value_type(node,to_be_multiplied));
		} else {
			DfgChain to_be_added;
			TedKids* kids = node->getKids();
			for(TedKids::iterator iterkid = kids->begin(), iterend = kids->end(); iterkid != iterend; iterkid++) {
				unsigned int index  = iterkid.getIndex();
				DfgNode* currentkid_dnode = current_dnode;	
				unsigned int current_dnode_delay = ceil(log10((float)(index))/log10((float)2))*DfgNode::delayMul();
				current_dnode_delay += (index>=1 && _respect_ntl_level) ? node->getVar()->getLevel() : 0;
				if (index>1) {
					currentkid_dnode = _dMapP.get(current_dnode,index);					
				} else if (index==0) {
					currentkid_dnode = _tMap.get(1);
				}

				TedNode* kid_node = iterkid.Node<TedNode>();
				wedge kid_weight = iterkid.getWeight();
				TedRegister kid_reg = iterkid.getRegister();
				unsigned int k = (kid_weight!=1 && kid_weight!=-1) ? 1 : 0;
				if (kid_node != TedNode::getOne()) {
					pair<bool,bool> tree_opkid = belong_to_chain(kid_node);
					bool retrieve_multiplicative_tree = tree_opkid.first;
					bool retrieve_additive_tree = tree_opkid.second;
					if (retrieve_additive_tree) {
						// inserts all additive tree to this node, the additive tree could be:
						// + forwarded if this node connects to an additive tree
						// + resolved at the for loop end if this node does not connect to an additive tree
						DfgChain adder_chain = delayAdditive[kid_node];
						if (connect_to_additive_tree) {
							to_be_added.insert(adder_chain.begin(),adder_chain.end());
							if (kid_weight==-1) {
								to_be_added.invert_add_tree();
							}
						} else {
							if (kid_reg!=0 || k!=0) {
								to_be_added.recompose_add_tree(_tMap.get(-1));
								to_be_added.create_add_tree();
								DfgNode* add_elem = to_be_added.begin()->second;
								long mask = ~0L -1;
								bool invert = (long)add_elem & 1;
								wedge weight_inv = kid_weight * (invert ? -1 : 1);
								unsigned int add_delay = to_be_added.begin()->first;
								if (weight_inv!=1 && weight_inv!=-1) {
									add_delay += DfgNode::delayMul();
								}
								add_elem = add_elem->mul(_tMap.get(weight_inv));
								if (kid_reg!=0) {
									add_elem = _dMapR.get(add_elem,kid_reg);
									add_delay = 0;
								}
								to_be_added.insert(DfgChain::value_type(add_delay,add_elem));
							} else {
								to_be_added.insert(adder_chain.begin(),adder_chain.end());
								if (kid_weight==-1) {
									to_be_added.invert_add_tree();
								}
							}
						}
					} else if (retrieve_multiplicative_tree) {
						//  + the multiplicative tree must be resolved right here, right now.
						//
						// re-compute latency and tree level
						DfgChain mul_chain = delayMultiplicative[kid_node];
						bool hasReg = (kid_reg!=0);
						unsigned int path_delay = 0;
						DfgNode* top_mul = NULL;
						if (hasReg) {
							mul_chain.create_mul_tree();
							top_mul = resolve_mul_tree(mul_chain,kid_weight,hasReg);
							top_mul = _dMapR.get(top_mul,kid_reg);
							top_mul = currentkid_dnode->mul(top_mul);
							path_delay = current_dnode_delay + DfgNode::delayMul();
						} else {
							mul_chain.insert(DfgChain::value_type(current_dnode_delay,currentkid_dnode));
							mul_chain.create_mul_tree();
							top_mul = resolve_mul_tree(mul_chain,kid_weight,hasReg);
							path_delay = mul_chain.begin()->first + 2*DfgNode::delayMul();

						}
						to_be_added.insert(DfgChain::value_type(path_delay,top_mul));
					} else {
						DfgNode* kid = _tMap.get(kid_node,1);
						unsigned int prior_latency = delayMap[kid_node];
						bool hasReg = (kid_reg!=0);
						unsigned int path_delay = 0;
						DfgNode* top_mul = NULL;
						if (hasReg) {
							//paths.insert(pair<unsigned int,TedNode*>(prior_latency,kid_node));
							top_mul = kid->mul(_tMap.get(kid_weight));
							top_mul = _dMapR.get(top_mul,kid_reg);
							top_mul = currentkid_dnode->mul(top_mul);
							path_delay = current_dnode_delay + DfgNode::delayMul();
						} else {
							bool invert = (kid_weight<0);
							kid_weight *= (invert) ? -1 : 1;
							if (kid_weight!=1) {
								if (prior_latency>current_dnode_delay) {
									top_mul = kid->mul(_tMap.get(kid_weight));
									top_mul = top_mul->mul(currentkid_dnode);
									path_delay = prior_latency + 2*DfgNode::delayMul();
								} else {
									top_mul = currentkid_dnode->mul(_tMap.get(kid_weight));
									top_mul = top_mul->mul(kid);
									path_delay = current_dnode_delay + 2*DfgNode::delayMul();
								}
							} else {
								top_mul = kid->mul(currentkid_dnode);
								path_delay = max(current_dnode_delay,prior_latency)+DfgNode::delayMul();
							}
							top_mul = reinterpret_cast<DfgNode*>((long)top_mul | invert);
						}
						to_be_added.insert(DfgChain::value_type(path_delay,top_mul));
					}
				} else {
					bool invert = (kid_weight<0);
					kid_weight *= (invert) ? -1 : 1;
					DfgNode* top_mul = currentkid_dnode;
					unsigned int path_delay = current_dnode_delay;
					if (kid_weight!=1) {
						top_mul = currentkid_dnode->mul(_tMap.get(kid_weight));
						path_delay += DfgNode::delayMul();
					}
					top_mul = reinterpret_cast<DfgNode*>( (long)top_mul | invert);
					to_be_added.insert(DfgChain::value_type(path_delay,top_mul));
				}
			}
			//
			// individual branches have been resolved
			// now we need to compute the collective delay of all
			// branches if the node does not require preservation
			//
			assert(!to_be_added.empty());
			if (connect_to_additive_tree) {
				assert(!_tMan->isPO(node));
				delayAdditive.insert(pair<TedNode*, DfgChain>(node,to_be_added));
			} else {
				if (to_be_added.size()>1) {
					to_be_added.recompose_add_tree(_tMap.get(-1));
					to_be_added.create_add_tree();
				}
				DfgNode* top_node = to_be_added.begin()->second;
				unsigned int top_delay = to_be_added.begin()->first;
				if ( (long)top_node & 1) {
					long mask = ~0L-1;
					top_node = reinterpret_cast<DfgNode*>( (long)top_node & mask);
					top_delay += DfgNode::delayMul();
					top_node = top_node->mul(_tMap.get(-1));
				}
				_tMap.registrate(node,1,top_node);
				delayMap[node] = top_delay;
			}
		}
	}

}

