/*
 * =====================================================================================
 *
 *       Filename:  DfgBitReduction.cc
 *    Description:
 *        Created:  11/04/2010 01:00:48 PM
 *         Author:  Daniel F. Gomez-Prado(dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision::                                              $: Revision of last commit
 *  $Author::                                                $: Author of last commit
 *  $Date::                                                  $: Date of last commit
 * =====================================================================================
 */

#include <cassert>
#include <cstring>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>
#include <climits>
#include <cmath>
using namespace std;

#include "TedOrderCost.h"
using namespace ted;

#include "DfgBitReduction.h"
#include "DfgNode.h"
#include "DfgMan.h"
#include "DfgOperator.h"

#include "util.h"
#include "csd.h"
using util::Util;
using util::Csd;

#include "Environment.h"
using tds::Environment;

namespace dfg {
	DfgBitReduction::DfgBitReduction(DfgMan* manager, double maxError, bool report)
		:_currentManager(manager), _maxError(maxError), _progressbar(report) {
		}

	void DfgBitReduction::opTopological(void) {
		//Goes left, right - top,bottom, decrementing the bitwidth by 1.
		DFGLevels wave;
		unsigned int latency = _currentManager->levelize(wave);
		DfgNode* lastnode;
		MinGappaBound gappa;
		FixCost cost;
		for(unsigned int index = 1; index <= latency; index++) {
			DFGSet& tochange = wave[index];
			DFGSet avoid;
			double error = 0.0;
			bool forceFinish = false;
			while((error < _maxError)&& !forceFinish) {
				int i = 1;
				for(set<DfgNode*>::iterator it = tochange.begin(); it != tochange.end(); it++,i++) {
					lastnode = (*it);
					if(avoid.find(*it) == avoid.end()) {
						if(lastnode->_op->getBitwidth()->isReducible()) {
							lastnode->_op->getBitwidth()->reduce();
							_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
							FixCost* costG = gappa.compute_cost(*_currentManager);
							double innerError  = costG->bitwidth;
							delete costG;
							error = innerError;
							if(innerError > _maxError) {
								lastnode->_op->getBitwidth()->increment();
								_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
								costG = gappa.compute_cost(*_currentManager);
								innerError = costG->bitwidth;
								error = innerError;
								delete costG;
								avoid.insert(lastnode);
							}
							if(_progressbar) {
								report(index,latency,lastnode,error);
							}
						} else {
							avoid.insert(lastnode);
						}
					}
				}
				forceFinish = (avoid.size() == tochange.size());
			}
		}
		if (_progressbar) {
			Util::progressBar(100,100,"Done. \"info -d\" for more information");
			cout << endl;
		}
	}

	void DfgBitReduction::report(unsigned int index, unsigned int max, DfgNode* node, double error) const {
		stringstream msg;
		msg << " op=" << node->getOpStr() << " id=" << node->_op->getID() << " error=" << error << " Q(" << node->_op->getBitwidth()->at() << ")";
		Util::progressBar(index,max,msg.str());
	}

	void DfgBitReduction::opTopologicalWithMulPriority(void) {
		//Goes left, right - top,bottom, decrementing the bitwidth by 1.
		DFGLevels wave;
		unsigned int latency = _currentManager->levelize(wave);
		DfgNode* lastnode;
		MinGappaBound gappa;
		FixCost cost;
		for(unsigned int index = 1; index <= latency; index++) {
			DFGSet& tochange = wave[index];
			DFGSet avoid;
			double error = 0.0;
			bool forceFinish = false;
			while((error < _maxError)&& !forceFinish) {
				int i = 1;
				int opCount = 0;
				for(DFGSet::iterator it = tochange.begin(); it != tochange.end(); it++,i++) {
					lastnode = (*it);
					if (lastnode->isAdd())
						continue;
					opCount++;
					if(avoid.find(*it) == avoid.end()) {
						if(lastnode->_op->getBitwidth()->isReducible()) {
							lastnode->_op->getBitwidth()->reduce();
							_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
							FixCost* costG = gappa.compute_cost(*_currentManager);
							double innerError  = costG->bitwidth;
							delete costG;
							error = innerError;
							if(innerError > _maxError) {
								lastnode->_op->getBitwidth()->increment();
								_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
								costG = gappa.compute_cost(*_currentManager);
								innerError = costG->bitwidth;
								error = innerError;
								delete costG;
								avoid.insert(lastnode);
							}
							if(_progressbar) {
								report(2*index,latency*2,lastnode,error);
							}
						} else {
							avoid.insert(lastnode);
						}
					}
				}
				forceFinish = (avoid.size() == opCount);
			}
			forceFinish = false;
			avoid.clear();
			while((error < _maxError)&& !forceFinish) {
				int i = 1;
				int opCount = 0;
				for(DFGSet::iterator it = tochange.begin(); it != tochange.end(); it++,i++) {
					lastnode = (*it);
					if (lastnode->isMul())
						continue;
					opCount++;
					if(avoid.find(*it) == avoid.end()) {
						if(lastnode->_op->getBitwidth()->isReducible()) {
							lastnode->_op->getBitwidth()->reduce();
							_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
							FixCost* costG = gappa.compute_cost(*_currentManager);
							double innerError  = costG->bitwidth;
							delete costG;
							error = innerError;
							if(innerError > _maxError) {
								lastnode->_op->getBitwidth()->increment();
								_currentManager->computeBitwidth(&DfgNode::uptoLevel,index);
								costG = gappa.compute_cost(*_currentManager);
								innerError = costG->bitwidth;
								error = innerError;
								delete costG;
								avoid.insert(lastnode);
							}
							if(_progressbar) {
								report(2*index+1,latency*2,lastnode,error);
							}
						} else {
							avoid.insert(lastnode);
						}
					}
				}
				forceFinish = (avoid.size() == opCount);
			}
		}
		if (_progressbar) {
			Util::progressBar(100,100,"Done. \"info -d\" for more information");
			cout << endl;
		}
	}

#if 0
	void DfgBitReduction::opPriorityCone(void) {
		//Goes left, right - top,bottom, decrementing the bitwidth by 1.
		bool report = _progressbar;
		_progressbar = false;
		opSensitivity();
		_progressbar = report;
		DfgParents parents;
		_currentManager->collectParents(parents);
		//DfgNode* lastnode;
		MinGappaBound gappa;
		FixCost cost;
		double error = 0.0;
		Mark<DfgNode>::newMark();
		for(Sensitivity::iterator it = _mpySense.begin(); it != _mpySense.end(); it++) {
			Cost opcost = it->first;
			DfgNode* node = it->second;
			unsigned int bitw = opcost.bitw;
			unsigned int diff = node->_op->getBitwidth()->getSize();

			//attempt full sensitivity reduction
			node->_op->getBitwidth()->reduce(bitw);

			//correct the bitwidth by binary approximation
			FixCost* costG = gappa.compute_cost(*_currentManager);
			error  = costG->bitwidth;
			delete costG;
			bool binaryCorrection = (error>_maxError);
			while(binaryCorrection) {
				bitw = bitw/2;
				if(error>_maxError) {
					node->_op->getBitwidth()->increment(bitw);
				} else {
					node->_op->getBitwidth()->reduce(bitw);
				}
				costG = gappa.compute_cost(*_currentManager);
				error  = costG->bitwidth;
				delete costG;
				binaryCorrection = (bitw>1) ? binaryCorrection : false;
				if(_progressbar) {
					cout << " error=" << error;
					cout << " op=" << node->getOpStr();
					cout << " id=" << node->_op->getID();
					cout << " bitwidth=" << node->_op->getBitwidth()->at()<< "       ";
					flush(cout);
					cout << '\xd';
				}
			}
			while (error>_maxError) {
				node->getOperator()->getBitwidth()->increment();
				costG = gappa.compute_cost(*_currentManager);
				error = costG->bitwidth;
				delete costG;
			}
			assert(error<=_maxError);
			if(_progressbar) {
				cout << " error=" << error;
				cout << " op=" << node->getOpStr();
				cout << " id=" << node->_op->getID();
				cout << " bitwidth=" << node->_op->getBitwidth()->at();
				flush(cout);
				cout << endl;
			}
			assert(node->_op->getBitwidth()->getSize()<=diff);
			diff -= node->_op->getBitwidth()->getSize();

			//propagate the exact bitwidth computation toward the outputs
			//this can be optimize if it is levelized
			DFGVector toUpdate = parents[node];
			while(!toUpdate.empty()) {
				DfgNode* update = toUpdate.back();
				toUpdate.pop_back();
				update->computeBitwidth();
				toUpdate.insert(toUpdate.begin(),parents[update].begin(),parents[update].end());
			}

			//propagate the exact bitwidth computation toward the inputs
			//inputs with different operators: preference is given to the path with multipliers
			//inputs with equal operators: bit reduction is split
			updateBitwidthTowardPI(node,diff);
		}
	}
#endif

	void DfgBitReduction::opPriority(void) {
		//Goes left, right - top,bottom, decrementing the bitwidth by 1.
		bool pb = _progressbar;
		_progressbar = false;
		opSensitivity();
		_progressbar = pb;
		MinGappaBound gappa;
		FixCost cost;
		double error = 0.0;
		Mark<DfgNode>::newMark();
		Sensitivity::size_type max = _mpySense.size() + _addSense.size();
		unsigned int count = 1;
		for (Sensitivity::reverse_iterator it = _mpySense.rbegin(); it != _mpySense.rend(); it++) {
			count++;
			Cost opcost = it->first;
			DfgNode* node = it->second;
			unsigned int bitw = opcost.bitw;
			unsigned int diff = node->_op->getBitwidth()->getSize();
			node->_op->getBitwidth()->reduce(bitw);
			FixCost* costG = gappa.compute_cost(*_currentManager);
			error  = costG->bitwidth;
			delete costG;
			while(error>_maxError) {
				node->_op->getBitwidth()->increment();
				costG = gappa.compute_cost(*_currentManager);
				error  = costG->bitwidth;
				delete costG;
				if(_progressbar) {
					report(count,max,node,error);
				}
			}
			assert(error<=_maxError);
			if(_progressbar) {
				report(count,max,node,error);
			}
			assert(node->_op->getBitwidth()->getSize()<=diff);
		}
		for (Sensitivity::reverse_iterator it = _addSense.rbegin() ; it != _addSense.rend(); it++) {
			count++;
			Cost opcost = it->first;
			DfgNode* node = it->second;
			unsigned int bitw = opcost.bitw;
			unsigned int diff = node->_op->getBitwidth()->getSize();
			node->_op->getBitwidth()->reduce(bitw);
			FixCost* costG = gappa.compute_cost(*_currentManager);
			error  = costG->bitwidth;
			delete costG;
			while(error>_maxError) {
				node->_op->getBitwidth()->increment();
				costG = gappa.compute_cost(*_currentManager);
				error  = costG->bitwidth;
				delete costG;
				if(_progressbar) {
					report(count,max,node,error);
				}
			}
			assert(error<=_maxError);
			if(_progressbar) {
				report(count,max,node,error);
			}
			assert(node->_op->getBitwidth()->getSize()<=diff);
		}
		if (_progressbar) {
			Util::progressBar(100,100,"Done. \"info -d\" for more information");
			cout << endl;
		}

	}

	void DfgBitReduction::opPriorityCone(void) {
		//Goes left, right - top,bottom, decrementing the bitwidth by 1.
		bool report = _progressbar;
		_progressbar = false;
		opSensitivity();
		_progressbar = report;
		DfgParents parents;
		_currentManager->collectParents(parents);
		//DfgNode* lastnode;
		MinGappaBound gappa;
		FixCost cost;
		double error = 0.0;
		Mark<DfgNode>::newMark();
		while ( !_mpySense.empty() ) {
			Sensitivity::iterator it = _mpySense.begin();
			Cost opcost = it->first;
			DfgNode* node = it->second;
			_mpySense.erase(it);
			unsigned int bitw = opcost.bitw;
			unsigned int diff = node->_op->getBitwidth()->getSize();

			//attempt full sensitivity reduction
			node->_op->getBitwidth()->reduce(bitw);

			//correct the bitwidth by binary approximation
			FixCost* costG = gappa.compute_cost(*_currentManager);
			error  = costG->bitwidth;
			delete costG;
			while(error>_maxError) {
				node->_op->getBitwidth()->increment();
				costG = gappa.compute_cost(*_currentManager);
				error  = costG->bitwidth;
				delete costG;
				if(_progressbar) {
					cout << " error=" << error;
					cout << " op=" << node->getOpStr();
					cout << " id=" << node->_op->getID();
					cout << " bitwidth=" << node->_op->getBitwidth()->at()<< "       ";
					flush(cout);
					cout << '\xd';
				}
			}
			assert(error<=_maxError);
			if(_progressbar) {
				cout << " error=" << error;
				cout << " op=" << node->getOpStr();
				cout << " id=" << node->_op->getID();
				cout << " bitwidth=" << node->_op->getBitwidth()->at();
				flush(cout);
				cout << endl;
			}
			assert(node->_op->getBitwidth()->getSize()<=diff);
			diff -= node->_op->getBitwidth()->getSize();

			//propagate the eccurate bitwidth computation toward the outputs
			//this can be optimize if it is levelized
			DFGVector toUpdate = parents[node];
			while(!toUpdate.empty()) {
				DfgNode* update = toUpdate.back();
				toUpdate.pop_back();
				update->recomputeBitwidth();
				toUpdate.insert(toUpdate.begin(),parents[update].begin(),parents[update].end());
			}

			//propagate the exact bitwidth computation toward the inputs
			//inputs with different operators: preference is given to the path with multipliers
			//inputs with equal operators: bit reduction is split
			updateBitwidthTowardPI(node,diff);
		}
	}


	void DfgBitReduction::updateBitwidthTowardPI(DfgNode* node,unsigned int val) {
		if(!node->isOp()|| val==0) {
			return;
		}
		if(node->isReg()) {
			return;
		}
		assert(node->getOp());
		MinGappaBound gappa;
		FixCost* costG = gappa.compute_cost(*_currentManager);
		double errorBound = costG->bitwidth;
		double errorG;
		delete costG;

		unsigned int lb = node->getLeft()->getOperator()->getBitwidth()->getSize();
		unsigned int rb = node->getRight()->getOperator()->getBitwidth()->getSize();
		if(node->isAddorSub()) {
			DfgNode* nb; //big
			DfgNode* ns; //small
			if(lb>rb) {
				nb = node->getLeft();
				ns = node->getRight();
			} else {
				ns = node->getLeft();
				nb = node->getRight();
			}
			if(nb->isOp()) {
				nb->getOperator()->getBitwidth()->reduce(val);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					nb->getOperator()->getBitwidth()->increment(val);
				} else {
					updateBitwidthTowardPI(nb,val);
				}
			}
			unsigned int n = nb->getOperator()->getBitwidth()->getSize();
			unsigned int m = ns->getOperator()->getBitwidth()->getSize();
			if(m>n && ns->isOp()) {
				ns->getOperator()->getBitwidth()->reduce(m-n);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					ns->getOperator()->getBitwidth()->increment(m-n);
				} else {
					updateBitwidthTowardPI(ns,m-n);
				}
			}
		} else {
			assert(node->isMul());
			if((node->getLeft()->isMul()&& node->getRight()->isMul())||
			   (node->getLeft()->isAddorSub()&& node->getRight()->isAddorSub())) {
				unsigned int l = val-val/2;
				unsigned int r = val/2;
				//unsigned int n;
				DfgNode* nl =  node->getLeft();
				DfgNode* nr = node->getRight();
				nl->getOperator()->getBitwidth()->reduce(l);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					nl->getOperator()->getBitwidth()->increment(l);
				} else {
					updateBitwidthTowardPI(nl,l);
				}
				nr->getOperator()->getBitwidth()->reduce(r);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					nr->getOperator()->getBitwidth()->increment(r);
				} else {
					updateBitwidthTowardPI(nr,r);
				}
			} else if(!node->getLeft()->isOp()&& !node->getLeft()->isOp()) {
				return;
			} else if(!node->getLeft()->isOp()|| !node->getLeft()->isOp()) {
				DfgNode* m = node->getLeft()->isOp() ? node->getLeft(): node->getRight();
				m->getOperator()->getBitwidth()->reduce(val);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					m->getOperator()->getBitwidth()->increment(val);
				} else {
					updateBitwidthTowardPI(m,val);
				}
			} else {
				DfgNode* m = node->getLeft()->isMul() ? node->getLeft(): node->getRight();
				assert(m->isMul());
				m->getOperator()->getBitwidth()->reduce(val);
				costG = gappa.compute_cost(*_currentManager);
				errorG = costG->bitwidth;
				delete costG;
				if( errorG > errorBound) {
					//do not accept change
					m->getOperator()->getBitwidth()->increment(val);
				} else {
					updateBitwidthTowardPI(m,val);
				}
			}
		}
	}



	/** @brief collects sensitivity information on all Arithmetic DFG operators
	 *  This method traverses from PIs to POs each level and changes the bitwidth
	 *  at each node to determine what is the maximum */
	void DfgBitReduction::opSensitivity(void) {
		DFGLevels wave;
		unsigned int latency = _currentManager->levelize(wave);
		DfgNode* lastnode;
		MinGappaBound gappa;
		FixCost* costG = gappa.compute_cost(*_currentManager);
		double initError  = costG->bitwidth;
		delete costG;
		if(_progressbar) {
			cout << "Error:  Initial=" << initError << "  maximum allow=" <<  _maxError << endl;
		}
		Cost opcost;
		for(unsigned int index = 1; index <= latency; index++) {
			DFGSet& tochange = wave[index];
			int i = 1;
			for(DFGSet::iterator it = tochange.begin(); it != tochange.end(); it++,i++) {
				lastnode = (*it);
				unsigned int bitw = lastnode->_op->getBitwidth()->getSize();
				double error = 0.0;
				bool goOn = true;
				while((error < _maxError)&& goOn) {
					if(lastnode->_op->getBitwidth()->isReducible()) {
						lastnode->_op->getBitwidth()->reduce();
						costG = gappa.compute_cost(*_currentManager);
						error  = costG->bitwidth;
						delete costG;
						if (error>_maxError) {
							// its better to backtrack here so that
							// the error information printed in the report
							// is accurate
							lastnode->_op->getBitwidth()->increment();
							costG = gappa.compute_cost(*_currentManager);
							error  = costG->bitwidth;
							delete costG;
							goOn = false;
						}
						if(_progressbar) {
							cout << "[" << index << "/" << latency << "]";
							cout << " " << i << "/" << tochange.size();
							cout << " error=" << error;
							cout << " op=" << lastnode->getOpStr();
							cout << " id=" << lastnode->_op->getID();
							cout << " bitwidth=" << lastnode->_op->getBitwidth()->at()<< "       ";
							flush(cout);
							cout << '\xd';
						}
					} else {
						goOn = false;
					}
				}
				assert(error<=_maxError);
				opcost.bitw = bitw - lastnode->_op->getBitwidth()->getSize();
				opcost.slope = (error-initError)/opcost.bitw;
				opcost.err = _maxError-error;
				if(_progressbar) {
					cout << endl;
				}
				if(lastnode->isMul()) {
					_mpySense.insert(pair<Cost,DfgNode*>(opcost,lastnode));
				} else {
					_addSense.insert(pair<Cost,DfgNode*>(opcost,lastnode));
				}
				lastnode->_op->getBitwidth()->increment(opcost.bitw);
			}
		}
		if(_progressbar) {
			cout << "Sensitivity cost report:" << endl;
			for(Sensitivity::iterator it = _mpySense.begin(); it != _mpySense.end(); it++) {
				opcost = it->first;
				lastnode = it->second;
				cout << lastnode->getOpStr()<< "(id=" <<  lastnode->_op->getID()<< ")";
				cout << "  slope=" << opcost.slope;
				cout << "  bit:reduction=" << opcost.bitw << "  error:inc=" << opcost.slope*opcost.bitw;
				cout << "  error:left=" << opcost.err << endl;
			}
			for(Sensitivity::iterator it = _addSense.begin(); it != _addSense.end(); it++) {
				opcost = it->first;
				lastnode = it->second;
				cout << lastnode->getOpStr()<< "(id=" <<  lastnode->_op->getID()<< ")";
				cout << "  slope=" << opcost.slope;
				cout << "  bit:reduction=" << opcost.bitw << "  error:inc=" << opcost.slope*opcost.bitw;
				cout << "  error:left=" << opcost.err << endl;
			}
		}
	}

}
