/*
 * =====================================================================================
 *
 *       Filename:  DfgMan.c
 *    Description:  DfgMan class
 *        Created:  05/01/2007 04:30:18 AM EDT
 *         Author:  Daniel Gomez-Prado(from jan 2008), Qian Ren(up to dec 2007)
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 208                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */


#include <cassert>
#include <cstring>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>
#include <climits>
#include <cmath>
using namespace std;

#include "TedOrderCost.h"
using namespace ted;

#include "DfgMan.h"
#include "DfgNode.h"
#include "DfgShifter.h"
#include "DfgOperator.h"

#include "util.h"
#include "csd.h"
using util::Util;
using util::Csd;

#include "Environment.h"
using tds::Environment;

namespace dfg {
	/**
	 * @brief DfgMan::writeCDFG writes the DFG graph in the CDFG format required by GAUT.
	 * @param file the name of the output file.
	 */
	void DfgMan::writeCDFG(const char* file) {
		//
		string word = Environment::getStr(Environment::BITWIDTH_INTEGER);
		//
		// use default bitwidth when creating the CDFG file unless the DFG has computed its own bitwidth
		//

		ofstream ofs(file);
		DFGVector vNodes, vConsts, vInputs, vOutputs, vVariables, vAgings;
		collectDFS(vNodes);
		ofs << "#VLSICAD-UMASS, 2012"<<endl;
		ofs << "#Date : " <<endl;
		ofs << "#Entry : poly2dfg" <<endl;
		ofs << "#  DFG output name -> DFG node pointer"<<endl;
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			ofs << "#\tNode_"<<p->first << " -> Node_" <<p->second << "\n";
		}
		ofs << "#"<<endl;
		for(DFGVector::size_type i = 0; i < vNodes.size(); i++) {
			if(vNodes[i]->isPI()) {
				vInputs.push_back(vNodes[i]);
			} else if (vNodes[i]->isReg()) {
				vAgings.push_back(vNodes[i]);
			} else if(vNodes[i]->isConst() || vNodes[i]->isVarConst()) {
				vConsts.push_back(vNodes[i]);
			} else {
				vVariables.push_back(vNodes[i]);
			}
		}
		ofs << "source(poly2dfg_start) {"<<endl;
		ofs << "\ttargets ";
		for(DFGVector::size_type i = 0; i < vConsts.size(); i++) {
			ofs <<(i==0?" ":",\n\t");
			if (vConsts[i]->isConst()) {
				ofs << Environment::getStr("const_prefix")<< Util::conformName(vConsts[i]->getConst());
			} else if (vConsts[i]->isVarConst()) {
				ofs << vConsts[i]->getName();
			} else {
				assert(false);
			}
		}
		if(!vConsts.empty()&& (!vInputs.empty() || !vAgings.empty()))
			ofs << ",\n\t";
		for(DFGVector::size_type i = 0; i < vInputs.size(); i++) {
			ofs <<(i==0?" ":",\n\t")<< vInputs[i]->getName();
		}
		if (!vInputs.empty() && !vAgings.empty()) {
			ofs << ",\n\t";
		}
		for(DFGVector::size_type i = 0; i < vAgings.size(); i++) {
			ofs <<(i==0?" ":",\n\t")<< writeCDFGop(vAgings[i]);
		}
		ofs << ";\n}" << endl;
		ofs << "#Constant declaration\n";
		for(DFGVector::size_type i = 0; i < vConsts.size(); i++) {
			DfgNode* const_i = vConsts[i];
			ofs << "constant(";
			if (const_i->isConst()) {
				ofs << Environment::getStr("const_prefix")<<Util::conformName(const_i->getConst());
			} else if (const_i->isVarConst()) {
				ofs << const_i->getName();
			} else {
				assert(false);
			}
			ofs <<") {\n";
			ofs << "\tbitwidth " << ( (const_i->getOperator() && const_i->getOperator()->getBitwidth()) ? const_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
			ofs << "\tsigned 1;\n";
			if (const_i->isConst()) {
				ofs << "\tvalue "<< const_i->getConst()<<";\n";
			} else if (const_i->isVarConst()) {
				ofs << "\tvalue "<<Util::conformName(const_i->getConst())<<";\n";
			} else {
				assert(false);
			}
			ofs << "\tbank 0;\n";
			ofs << "\taddress "<< i <<";\n";
			ofs << "\thardwire 1;\n";
			ofs << "}\n";
		}
		ofs << "#Input declaration\n";
		for(DFGVector::size_type i = 0; i < vInputs.size(); i++) {
			DfgNode* input_i = vInputs[i];
			ofs << "input("<< getNameIfRegisters(input_i) <<") {\n";
			ofs << "\tbitwidth " << ( (input_i->getOperator() && input_i->getOperator()->getBitwidth()) ? input_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
			ofs << "\tsigned 1;\n";
			ofs << "}\n";
		}
		ofs << "#Output declaration\n";
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			DfgNode* output_i = p->second;
			ofs << "output(" << p->first << ") {\n";
			ofs << "\tbitwidth " << ( (output_i->getOperator() && output_i->getOperator()->getBitwidth()) ? output_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
			ofs << "\tsigned 1;\n";
			ofs << "}\n";
		}

		ofs << "#Aging declaration\n";
		for(DFGVector::size_type i = 0; i < vAgings.size(); i++) {
			DfgNode* aging_i = vAgings[i];
			ofs << "variable("<< writeCDFGop(aging_i) <<") {\n";
			if (aging_i->getLeft()->isOp()) {
				ofs << "\taging " << writeCDFGop(aging_i->getLeft()) << ";\n";
			} else {
				ofs << "\taging " << getNameIfRegisters(aging_i->getLeft()) << ";\n";
			}
			ofs << "\tbitwidth " << ( (aging_i->getOperator() && aging_i->getOperator()->getBitwidth()) ? aging_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
			ofs << "\tsigned 1;\n";
			ofs << "}\n";
		}
		ofs << "#Variable declaration\n";
		unsigned int eq_loopback_out = 0;
		for(DFGVector::size_type i = 0; i < vVariables.size(); i++) {
			DfgNode* var_i = vVariables[i];
			if (var_i->getOp() == DfgOperator::REG) {
				ofs << "# Registers #\n";
				assert(var_i->getLeft() ==var_i->getRight());
				wedge start = var_i->getRegisterValue();
				wedge RCount = start;
				for (;RCount>0;RCount--) {
					ofs << "variable("<< writeCDFGop(var_i);
					if (RCount!=start) {
						ofs << "R" << RCount;
					}
					ofs << ") {\n";
					if (RCount!=1) {
						ofs << "\taging " << writeCDFGop(var_i) << "R" << RCount-1 << ";\n";
					} else {
						ofs << "\taging ";
						writeCDFGname(ofs,var_i->getLeft());
						ofs << ";\n";
					}
					ofs << "\tbitwidth " << ( (var_i->getOperator() && var_i->getOperator()->getBitwidth()) ? var_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
					ofs << "\tsigned 1;\n";
					ofs << "}\n";
				}
			} else if (var_i->getOp() == DfgOperator::EQ || var_i->getOp() ==DfgOperator::LOOPBACK) {
				// we are assuming EQ and LOOPBACK are only present as PO nodes and therefore
				// we can skip them and recover them at the output nodes.
				eq_loopback_out++;
				continue;
			} else {
				ofs << "variable("<<writeCDFGop(var_i)<<") {\n";
				ofs << "\tbitwidth " << ( (var_i->getOperator() && var_i->getOperator()->getBitwidth()) ? var_i->getOperator()->getBitwidth()->at() : word ) << ";\n";
				ofs << "\tsigned 1;\n";
				ofs << "}\n";
			}
		}
		ofs <<"#Operations\n";
		for(DFGVector::size_type i = 0; i < vVariables.size(); i++) {
			DfgNode* var_i = vVariables[i];
			if(var_i->isOnechild()) {
				//REG ASSIGN LOOPBACK
				continue;
			}
			string func = DfgOperator::getOpNameCDFG(var_i->getOp());
			ofs << "operation(dfg_op"<<i<<") {\n";
			ofs << "\tfunction " << func <<";\n";
			ofs << "\tread ";
			writeCDFGname(ofs,var_i->getLeft());
			ofs << ",";
			writeCDFGname(ofs,var_i->getRight());
			ofs << ";\n";
			DfgMan::PrimaryOutputs::iterator p = _mPos.begin(), p_end = _mPos.end();
			bool found = false;
			while (p!=p_end & !found) {
				if (p->second == var_i) {
					found = true;
				} else {
					p++;
				}
			}
			if (found) {
				ofs << "\twrite " << p->first << ";\n";
			} else  {
				ofs << "\twrite "<< writeCDFGop(var_i)<<";\n";
			}
			ofs << "}\n";
		}
		ofs << "#assign to outputs" << endl;
		DFGVector::size_type j = vVariables.size();
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			if (p->second->getOp() == DfgOperator::EQ || p->second->getOp() ==DfgOperator::LOOPBACK) {
				ofs << "operation(op"<<j++<<") {\n";
				eq_loopback_out--;
				string func = DfgOperator::getOpNameCDFG(p->second->getOp());
				ofs << "\t#stitching back" << endl;
				ofs << "\tfunction " << func << ";" << endl;
				ofs << "\tread ";
				writeCDFGname(ofs,p->second->getLeft());
				ofs << ";\n";
				ofs << "\twrite " << p->first << ";\n";
				ofs << "}\n";
			} else {
#if 0
				// do not add assigns as it adds overhead

				// a normal operation to which we are appending the assign function
				ofs << "operation(op"<<j++<<") {\n";
				ofs << "\t#adding output" << endl;
				ofs << "\tfunction assign;" << endl;
				ofs << "\tread " << writeCDFGop(p->second) << ";\n";
				ofs << "\twrite " << p->first << ";\n";
				ofs << "}\n";
#endif
			}
		}
		assert(eq_loopback_out==0);
		ofs <<"sink(poly2dfg_end) {\n";
		ofs <<"\ttargets ";
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			if(p == _mPos.begin())
				ofs << "\t" << p->first;
			else
				ofs << ",\n\t" << p->first;
		}
		ofs <<";\n";
		ofs <<"}";
		ofs.close();

		//
		// generate constraints
		//
		string constraint_file = file;
		constraint_file = Util::replaceFileExtension(file,"ioc");
		ofstream iocfs(constraint_file.c_str());
		assert(iocfs.is_open());
		unsigned int port = 1;
		iocfs << "#VLSICAD-UMASS, 2012"<<endl;
		iocfs << "TABLE " << vInputs.size()+_mPos.size() << " 4" << endl;
		for(DFGVector::size_type i = 0; i < vInputs.size(); i++) {
			iocfs << vInputs[i]->getName() << " Input " << port++ << " 0" << endl;
		}
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			iocfs << p->first << " Output " << port++ << " " << Environment::getInt(Environment::GAUT_CADENCY)-Environment::getInt(Environment::GAUT_CLOCK) << endl;
		}
		iocfs.close();
	}

	void DfgMan::writeCDFGname(ofstream& ofs, DfgNode* pNode) {
		switch(pNode->getOp()) {
		case DfgOperator::CONST:
			ofs << Environment::getStr("const_prefix")<< Util::conformName(pNode->getConst());
			break;
		case DfgOperator::VARCONST:
		case DfgOperator::VAR:
			ofs << getNameIfRegisters(pNode);
			break;
		default:
			ofs << writeCDFGop(pNode);
			break;
		}
	}

	string DfgMan::writeCDFGop(DfgNode* pNode) {
		string name = "id";
		name += Util::itoa(pNode->getOperator()->getID());
		return name;
	}

}
