/*
 * =====================================================================================
 *
 *       Filename:  DfgMan.c
 *    Description:  DfgMan class
 *        Created:  05/01/2007 04:30:18 AM EDT
 *         Author:  Daniel Gomez-Prado(from jan 2008), Qian Ren(up to dec 2007)
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 208                                          $: Revision of last commit
 *  $Author:: daniel@win                                     $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2010)$: Date of last commit
 * =====================================================================================
 */


#include <cassert>
#include <cstring>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>
#include <climits>
#include <cmath>
using namespace std;

#include "TedOrderCost.h"
using namespace ted;

#include "DfgMan.h"
#include "DfgNode.h"
#include "DfgShifter.h"
#include "DfgOperator.h"

#include "util.h"
#include "csd.h"
using util::Util;
using util::Csd;

#include "Environment.h"
using tds::Environment;

namespace dfg {

	/**
	 * @brief DfgMan::showDot generates the DFG dot file
	 *
	 * @param bUseArrivalTime
	 * @param bVerbose
	 **/
	void DfgMan::writeDot(ofstream& ofile, bool bUseArrivalTime, bool bVerbose, bool show_ids) {
		unsigned int nMul = 0;
		unsigned int nAdd = 0;
		DFGVector vNodes;
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n"
			<< "\t\tnode [shape = plaintext];\n"
			<< "\t\tedge [style = invis];\n"
			<< "\t\tLevelPO [label=\"\"];\n";
		unsigned int maxLevel = collectDFS(vNodes);
		if(bUseArrivalTime) {
			maxLevel = maxArrivalTime();
		}
		for(unsigned int i = 0; i <= maxLevel; i++) {
			ofile << "\t\tLevel" << i << " [label=\"";
			if(Environment::getBool("show_level"))
				ofile << i;
			ofile << "\"];\n";
		}
		ofile << "\t\tLevelComments [label=\"\"];\n";
		for(unsigned int i = 0; i <= maxLevel; i++) {
			ofile << "Level" << i << " -> ";
		}
		ofile << "\t\tLevelPO";
		ofile << "-> LevelComments";
		ofile << "\n\t}\n";
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			ofile << "\t{rank = same; LevelPO; Node_" << p->first;
			ofile << " [label=\""<<p->first<<"\" shape=triangle, fontsize=26]; }\n";
			ofile << "\tNode_"<<p->second << " -> Node_" <<p->first << "\n";
		}
		for(DFGVector::size_type i = 0; i < vNodes.size(); i ++) {
			if(bUseArrivalTime) {
				ofile << dotString(vNodes[i], vNodes[i]->getArrivalTime(), bVerbose, show_ids);
			} else {
				ofile << dotString(vNodes[i], vNodes[i]->getLevel(), bVerbose, show_ids);
			}
			if(vNodes[i]->isMul())nMul++;
			if(vNodes[i]->isAdd()|| vNodes[i]->isSub())nAdd++;
		}
		ofile << "\t{rank=same;LevelComments;text [shape=plaintext,font=40, label=\"";
		ofile << "ADD(s) = " << nAdd << " MPY(s) = " << nMul <<" Latency = "<< maxLevel;
		ofile << "\"]}\n";
		ofile << "}";
	}



	string DfgMan::dotString(DfgNode* pNode, unsigned int level, bool bVerbose, bool show_ids) {
		ostringstream oss;
		if(pNode->_op && pNode->_op->getBitwidth()) {
			oss << "\n\tsubgraph bitw_" << pNode << "{";
			oss << "\n\trank = same; Level" << level <<"; node_bitw_" << pNode << "[label=\"";
			if (pNode->_op && pNode->_op->getBitwidth()) {
				oss << "Q(" << pNode->_op->getBitwidth()->at() << ")";
			}
			oss << "\", shape = plaintext]";
			oss << "\n\tNode_" << pNode << " -> node_bitw_" << pNode << "[arrowhead=inv, minlen=0]";
		}
		oss << "\n\t{rank = same; Level"<<level<<"; Node_"<<pNode<<" [label=\"";
		switch(pNode->getOp()) {
		case DfgOperator::CONST:
			oss << pNode->getConst();
			break;
		case DfgOperator::VAR:
			{
				oss << pNode->getName() << " ";
				//oss << getNameIfRegisters(pNode);
				break;
			}
		case DfgOperator::VARCONST:
			{
				TedVar* pvar = pNode->getTedVarIfExist();
				if(pvar && pvar->hasRange())
					oss << pvar->getRange();
				else
					oss << pNode->getName();
				break;
			}
		default:
			{
				oss << pNode->getOpStr();
			}
		}
		if (show_ids) {
			oss << "[" << pNode->getOperator()->getID() << "]";
		}
		if(bVerbose) {
			oss << " \\["<<pNode->getArrivalTime()<<"/" <<pNode->getRequiredTime()<<"\\]";
		}
		if(pNode->isInput()) {
			oss <<"\" shape=invtriangle fontsize= 26]; }\n";
		} else  {
			if(pNode->isAddorSub()) {
				oss <<"\" shape= box fontsize= 26]; }\n";
			} else if(pNode->isOnechild()) {
				oss << "\" shape=box height=0.25 fontsize=18]; }\n";
			} else { // all other operations
				oss <<"\" shape=ellipse fontsize= 26]; }\n";
			}
		}
		if(pNode->_op && pNode->_op->getBitwidth()) {
			oss << "\n\t}";
		}
		if(!pNode->isInput()) {
			oss <<"\tNode_"<<pNode->getLeft()<<" -> Node_"<<pNode;
			oss <<" [style=solid];\n";
			if(!pNode->isOnechild()) {
				oss <<"\tNode_"<<pNode->getRight()<<" -> Node_"<<pNode;
				oss <<" [style="<<(pNode->isLRsensible() ?"dashed":"solid")<<"];\n";
			}
		}
		return oss.str();
	}

	void DfgMan::writeDotDetailed(void) {
		static unsigned int index = 0;
		ofstream ofile;
		string filename = Environment::getStr("show_directory")+ "/debugDfg" + Util::itoa(index++)+ ".dot";
		ofile.open(filename.c_str());
		if(!ofile.is_open()) {
			throw(string("01012. Cannot open temporary file for writing.\n"));
		}
		writeDotDetailed(ofile);
		ofile.close();
		Util::showDotFile(filename.c_str());
	}

	void DfgMan::writeDotDetailed(ofstream& ofile) {
		unsigned int nMul = 0;
		unsigned int nAdd = 0;
		vector <DfgNode* > vNodes;
		ofile << "Digraph G {\n\tsize=\"10, 7.5\"; center = \"true\"; \n";
		ofile << "\t{\n"
			<< "\t\tnode [shape = plaintext];\n"
			<< "\t\tedge [style = invis];\n"
			<< "\t\tLevelPO [label=\"\"];\n";
		//collectDFSonly(vNodes);
		int maxLevel = maxArrivalTime();
		for(int i = maxLevel; i >= 0; i--) {
			ofile << "\t\tLevel" << i << " [label=\""<< i <<"\"];\n";
		}
		ofile << "\t\tLevelComments [label=\"\"];\n";
		for(int i = 0; i <= maxLevel; i++) {
			ofile << "Level" << i << " -> ";
		}
		ofile << "\t\tLevelPO";
		ofile << "-> LevelComments";
		ofile << "\n\t}\n";
		for(DfgMan::PrimaryOutputs::iterator p = _mPos.begin(); p != _mPos.end(); p++) {
			ofile << "\t{rank = same; LevelPO; Node_" << p->first;
			ofile << " [label=\""<<p->first<<"\" shape=triangle, fontsize=26]; }\n";
			ofile << "\tNode_"<<p->second << " -> Node_" <<p->first << "\n";
		}
		for(set<DfgNode*>::const_iterator it = _sNodes.begin(), end = _sNodes.end(); it != end; it++) {
			//	if((*it)->_visited.isMarked()) {
			ofile << dotStringDetailed(*it);
			if((*it)->isMul())nMul++;
			if((*it)->isAdd()||(*it)->isSub())nAdd++;
			//	}
		}
		ofile << "\t{rank=same;LevelComments;text [shape=plaintext,font=40, label=\"";
		ofile << "ADD(s) = " << nAdd << " MPY(s) = " << nMul <<" Latency = "<< maxLevel;
		ofile << "\"]}\n";
		ofile << "}";
	}

	string DfgMan::dotStringDetailed(DfgNode* pNode) {
		ostringstream oss;
		oss << "\n\t{Node_"<<pNode<<" [label=\"";
		switch(pNode->getOp()) {
		case DfgOperator::CONST:
			oss << pNode->getConst();
			break;
		case DfgOperator::VAR:
			oss << pNode->getName();
			break;
		case DfgOperator::VARCONST:
			{
				TedVar* pvar = pNode->getTedVarIfExist();
				if(pvar && pvar->hasRange())
					oss << pvar->getRange();
				else
					oss << pNode->getName();
				break;
			}
		case DfgOperator::REG:
		case DfgOperator::ADD:
		case DfgOperator::SUB:
		case DfgOperator::MUL:
		case DfgOperator::DIV:
		case DfgOperator::LSH:
		case DfgOperator::RSH:
		case DfgOperator::EQ:
			oss << pNode->getOpStr();
			if(pNode->_data.ptr)
				oss << "(ptr=" << pNode->_data.ptr << ")";
			oss << " [" << pNode->getOperator()->getID() << "]";
			break;
		default:
			throw(string("01022. UNKNOWN DFG node type."));
			break;
		}
		oss << " \\n" << pNode;
		oss << " \\n\\["<<pNode->getArrivalTime()<<"/" <<pNode->getRequiredTime()<<"\\]";
		oss << " \\n l=" << pNode->_level << " v=" << pNode->_value;
		oss <<(pNode->_visited.isMarked() ?" M":" U")<< pNode->_visited.getMark();
		if (pNode->_op && pNode->_op->getBitwidth()) {
			oss << " \\n Q(" << pNode->_op->getBitwidth()->at() << ")";
		}
		if(pNode->isOp()) {
			if(pNode->isAddorSub()) {
				oss <<"\" shape=box fontsize= 26]; }\n";
			} else if(pNode->isReg()) {
				oss << "\" shape=box height=0.25 fontsize=18]; }\n";
			} else {
				oss <<"\" shape= ellipse fontsize= 26]; }\n";
			}
		} else  {
			oss <<"\" shape=invtriangle fontsize= 26]; }\n";
		}
		if(pNode->isOp()) {
			oss <<"\tNode_"<<pNode->getRight()<<" -> Node_"<<pNode;
			oss <<" [style="<<((pNode->isSub()||pNode->isLSH()) ?"dashed":"solid")<<", label=\"R\"];\n";
			if(!pNode->isOnechild()) {
				oss <<"\tNode_"<<pNode->getLeft()<<" -> Node_"<<pNode;
				oss <<" [style=solid, label=\"L\"];\n";
			}
		}
		return oss.str();
	}

}
