/*
 * =====================================================================================
 *
 *       Filename:  XMLImport.h
 *    Description:
 *        Created:  7/4/2012 10:50:27 AM
 *         Author:  Daniel f. Gomez-Prado(dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 203                                          $: Revision of last commit
 *  $Author:: daniel                                         $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2012)$: Date of last commit
 * =====================================================================================
 */
#include <cassert>
#include <iostream>

#include "DfgMan.h"
#include "DfgOperator.h"
#include "DfgNode.h"

#if _WIN32
#include "../3rdparty/tinyxml2/tinyxml2.h"
#else
#include "tinyxml2.h"
#endif

using namespace tinyxml2;

namespace dfg {


	/** Records the Dfgs existing in the Dfg Manager*/
	void DfgMan::writeXML(string filename) {
		if (filename.empty()) {
			return;
		}
		const char* header = 
			"<?xml version=\"1.0\" encoding=\"ASCII\"?>"
			"<!-- XML content of the DFG data structure of TDS -->"
			"<tds.dfg.Graph tds.webpage=\"http://incascout.ecs.umass.edu/main\" >"
			"</tds.dfg.Graph>";
		XMLDocument* pxml = new XMLDocument();
		XMLDocument& xml = *pxml;
		xml.Parse(header);
		XMLElement* root = xml.RootElement();

		DFGVector vNodes;
		unsigned int maxLevel = collectDFS(vNodes);

		// Start inserting the node elements
		//not the optimal implementation, but it gets the job done
		unsigned int max_id = 0;
		for(unsigned int i = 0; i <= maxLevel; i ++) {
			string comment = string("DFG Level") +Util::itoa(i);
			XMLComment* level_comment = xml.NewComment(comment.c_str());
			root->InsertEndChild(level_comment);
			for(DFGVector::size_type j = 0; j < vNodes.size(); j ++) {
				DfgNode* pNode = vNodes[j];
				if(pNode->getLevel() ==i) {
					writeXML(xml,pNode);
					if (pNode->getOperator()->getID()>max_id) {
						max_id = pNode->getOperator()->getID();
					}
				}
			}
		}
		max_id++;
		// write all the outputs out
		root->InsertEndChild( xml.NewComment("Defining the DFG Primary Outputs"));
		vector<XMLElement*> out_edges;
		for(map<string, DfgNode*>::iterator p = _mPos.begin(); p != _mPos.end(); p++, max_id++) {
			XMLElement* element_node = xml.NewElement("node");
			element_node->SetAttribute("name",p->first.c_str());
			element_node->SetAttribute("id",max_id);
			root->InsertEndChild(element_node);

			XMLElement* element_attr1 = xml.NewElement("attr");
			element_attr1->SetAttribute("key","CLASS");
			element_attr1->SetAttribute("value","DATA");
			element_node->InsertEndChild(element_attr1);

			XMLElement* element_attr2 = xml.NewElement("attr");
			element_attr2->SetAttribute("key","NAME");
			element_attr2->SetAttribute("value",p->first.c_str());
			element_node->InsertEndChild(element_attr2);

			XMLElement* element_attr3 = xml.NewElement("attr");
			element_attr3->SetAttribute("key","NATURE");
			element_attr3->SetAttribute("value","VARIABLE_VAL_UNKNOWN");
			element_node->InsertEndChild(element_attr3);			

			XMLElement* element_attr4 = xml.NewElement("attr");
			element_attr4->SetAttribute("key","PROPERTY");
			element_attr4->SetAttribute("value","OUTPUT");
			element_node->InsertEndChild(element_attr4);

			XMLElement* element_edge = xml.NewElement("edge");
			element_edge->SetAttribute("id_pred",p->second->getOperator()->getID());
			element_edge->SetAttribute("id_succ",max_id);			
			out_edges.push_back(element_edge);
		}

		// write all the edges
		root->InsertEndChild( xml.NewComment("Interconnecting the DFG nodes"));
		for(DFGVector::size_type j = 0; j < vNodes.size(); j ++) {
			DfgNode* pNode = vNodes[j];
			if (pNode->isInput()) {
				continue;
			}
			DfgNode* left = pNode->getLeft();
			XMLElement* element_edge1 = xml.NewElement("edge");
			element_edge1->SetAttribute("id_pred",left->getOperator()->getID());
			element_edge1->SetAttribute("id_succ",pNode->getOperator()->getID());
			root->InsertEndChild(element_edge1);
			if (pNode->isOnechild()) {
				continue;
			}
			if (pNode->isLRsensible()) {
				XMLElement* element_attr1 = xml.NewElement("attr");
				element_attr1->SetAttribute("key","NUM_INPUT");
				element_attr1->SetAttribute("value","0");
				element_edge1->InsertEndChild(element_attr1);
			}
			
			DfgNode* right = pNode->getRight();
			XMLElement* element_edge2 = xml.NewElement("edge");
			element_edge2->SetAttribute("id_pred",right->getOperator()->getID());
			element_edge2->SetAttribute("id_succ",pNode->getOperator()->getID());
			root->InsertEndChild(element_edge2);
			if (pNode->isLRsensible()) {
				XMLElement* element_attr2 = xml.NewElement("attr");
				element_attr2->SetAttribute("key","NUM_INPUT");
				element_attr2->SetAttribute("value","1");
				element_edge2->InsertEndChild(element_attr2);
			}
		}
		for (vector<XMLElement*>::size_type index = 0; index < out_edges.size(); index++) {
			root->InsertEndChild(out_edges[index]);
		}

		xml.SaveFile(filename.c_str());
	}

	void DfgMan::writeXML(XMLDocument& xml, DfgNode* pNode) {
		XMLElement* root = xml.RootElement();

		XMLElement* element_node = xml.NewElement("node");
		element_node->SetAttribute("name",pNode->getOpStr());
		element_node->SetAttribute("id",pNode->getOperator()->getID());
		root->InsertEndChild(element_node);

		XMLElement* element_attr1 = xml.NewElement("attr");
		element_attr1->SetAttribute("key","CLASS");
		element_attr1->SetAttribute("value",pNode->isInput() ? "DATA" : "OP");
		element_node->InsertEndChild(element_attr1);

		XMLElement* element_attr3 = xml.NewElement("attr");
		element_attr3->SetAttribute("key","NATURE");
		element_attr3->SetAttribute("value",DfgOperator::getOpNameXML(pNode->getOp()));
		element_node->InsertEndChild(element_attr3);

		if (pNode->isInput()) {
			XMLElement* element_attr2 = xml.NewElement("attr");
			element_attr2->SetAttribute("key","NAME");
			element_attr2->SetAttribute("value",pNode->getOpStr());
			element_node->InsertEndChild(element_attr2);

			XMLElement* element_attr4 = xml.NewElement("attr");
			element_attr4->SetAttribute("key","PROPERTY");
			element_attr4->SetAttribute("value","INPUT");
			element_node->InsertEndChild(element_attr4);
		}
	}
}