/*
 * =====================================================================================
 *
 *       Filename:  XMLImport.h
 *    Description:
 *        Created:  7/4/2012 10:50:27 AM
 *         Author:  Daniel f. Gomez-Prado(dgomezpr), dgomezpr@ecs.umass.edu
 *        Company:  UMASS
 *
 * =====================================================================================
 *  $Revision:: 203                                          $: Revision of last commit
 *  $Author:: daniel                                         $: Author of last commit
 *  $Date:: 2010-11-18 08:13:53 -0500(Thu, 18 Nov 2012)$: Date of last commit
 * =====================================================================================
 */
#include <cassert>
#include <iostream>

#include "XMLImport.h"
#include "DfgOperator.h"
#include "DfgNode.h"


namespace dfg {

		const char* XMLImport::XML_NODE = "node";
		const char* XMLImport::XML_EDGE = "edge";
		const char* XMLImport::XML_ATTR = "attr";
		const char* XMLImport::XML_KEY = "key";
		const char* XMLImport::XML_VALUE = "value";

		const char* XMLImport::ATTR_CLASS = "CLASS";
		const char* XMLImport::ATTR_NAME = "NAME";
		const char* XMLImport::ATTR_NATURE = "NATURE";
		const char* XMLImport::ATTR_PROPERTY = "PROPERTY";
		const char* XMLImport::ATTR_NUM_INPUT = "NUM_INPUT";
		const char* XMLImport::ATTR_INDEXES = "INDEXES";
		const char* XMLImport::ATTR_VALUE = "VALUE";
		const char* XMLImport::ATTR_VAL_MIN = "VAL_MIN";
		const char* XMLImport::ATTR_VAL_MAX = "VAL_MAX";

		namespace {
			void generateName(string& name) {
				Util::strSubstitute(name,"[","_");
				Util::strSubstitute(name,"]","");
			}
		}

		bool XMLImport::VisitEnter( const XMLDocument& doc ) {
			return true;
		}

		bool XMLImport::VisitEnter( const XMLElement& element, const XMLAttribute* attribute ) {
			OpenElement( element.Name() );
				return 	ProcessElement(element);
		}


		bool XMLImport::VisitExit( const XMLElement& ) {
			CloseElement();
				return true;
		}

		void XMLImport::OpenElement( const char* name ) {
			stack.Push( name );
				++depth;
		}

		void XMLImport::CloseElement() {
			--depth;
				const char* name = stack.Pop();
		}

		bool XMLImport::ProcessElement(const XMLElement& element) {
			if(!strcmp(element.Name(),XML_NODE)) {
				ProcessNode(element);
					return false;
			} else if (!strcmp(element.Name(),XML_EDGE)) {
				ProcessEdge(element);
					return false;
			} else {
				return true;
			}
		}

		void XMLImport::ProcessNode(const XMLElement& element) {
			DfgOperator::Type nodeType = DfgOperator::UNKNOWN;
			const char* attrName = NULL;
			const char* attrTypeStr = NULL;
			const char* attrProperty = NULL;
			const char* attrIndexes = NULL;
			float minValue = 0.0;
			float maxValue = 0.0;
			float attrValue = 0.0;
			bool isVarConst = false;
			const char* nodeName = element.Attribute("name");
			int nodeID = element.IntAttribute("id");
			assert(nodeName);
			assert(nodeID>0);
			string bracket_name(Util::strsav(nodeName));
			bracket_name += "_";
			bracket_name += Util::strsav(element.Attribute("id"));
			string node_name = Util::strsav(bracket_name.c_str());
			generateName(node_name);
			for ( const XMLNode* node=element.FirstChild(); node; node=node->NextSibling() ) {
				const XMLElement* enode = node->ToElement();
				if (enode) {
					assert(!strcmp(enode->Name(),XML_ATTR));
					if (!strcmp(enode->Attribute(XML_KEY),ATTR_CLASS)) {
						const char* node_type = enode->Attribute(XML_VALUE);
						if (!strcmp(node_type,"DATA")) {
							nodeType = DfgOperator::VAR;
						} else {
							assert(!strcmp(node_type,"OP"));
						}
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_NAME)) {
						attrName = enode->Attribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_NATURE)) {
						attrTypeStr = enode->Attribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_PROPERTY)) {
						attrProperty = enode->Attribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_INDEXES)) {
						attrIndexes = enode->Attribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_VAL_MIN)) {
						minValue = enode->FloatAttribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_VAL_MAX)) {
						maxValue = enode->FloatAttribute(XML_VALUE);
					} else if (!strcmp(enode->Attribute(XML_KEY),ATTR_VALUE)) {
						attrValue = enode->FloatAttribute(XML_VALUE);
						isVarConst = true;
					}
				}
			}
			DfgNode* pNode = NULL;
			if (nodeType==DfgOperator::VAR) {
				if (isVarConst) {
					// refine selection
					nodeType = DfgOperator::VARCONST;
				}
				assert(attrName);
				DfgMapID::iterator IDIterator = IDTable.find(nodeID);
				if(IDIterator != IDTable.end()) {
					throw(string("09001. Duplicated unique ID for node ")+ node_name);
				}
				if (attrProperty && !strcmp(attrProperty,"OUTPUT")) {
					//XML file is malformed, an output should not be a constant variable
					assert(nodeType==DfgOperator::VAR);
					PoTable.insert(pair<int,string>(nodeID,node_name));
				} else {
					pNode = new DfgNode(_dMan,node_name,nodeType);
					_dMan->registerNode(pNode);
					IDTable.insert(pair<int,DfgNode*>(nodeID,pNode));
					if (attrIndexes) {
						LoopbackTable.insert(pair<int,bool>(nodeID,true));
					}
				}
			} else {
					// node has to be a OP
				DfgMapID::iterator IDIterator = IDTable.find(nodeID);
				assert(IDIterator==IDTable.end());
				nodeType = DfgOperator::queryType(attrTypeStr);
				if (nodeType==DfgOperator::REG) {
					string regamount = "1";
					regamount += DfgNode::TOKEN_REG;
					pNode = new DfgNode(_dMan,regamount,nodeType);
				} else {
					pNode = new DfgNode(_dMan,nodeType,NULL,NULL);
				}
				if (nodeType==DfgOperator::UNKNOWN) {
					std::cout << "Parse unsupported operator " << attrTypeStr << std::endl;
				}
				IDTable.insert(pair<int,DfgNode*>(nodeID,pNode));
			}
		}

		void XMLImport::ProcessEdge(const XMLElement& element) {
			int id_pred = element.IntAttribute("id_pred");
			int id_succ = element.IntAttribute("id_succ");
			const char* edgeW = NULL;
			for ( const XMLNode* node=element.FirstChild(); node; node=node->NextSibling() ) {
				const XMLElement* enode = node->ToElement();
				if (enode) {
					assert(!strcmp(enode->Name(),XML_ATTR));
					if (!strcmp(enode->Attribute(XML_KEY),ATTR_NUM_INPUT)) {
						edgeW = enode->Attribute(XML_VALUE);
					}
				}
			}
			assert(IDTable.find(id_pred)!=IDTable.end());
			DfgNode* pred = IDTable.find(id_pred)->second;
			if (IDTable.find(id_succ) ==IDTable.end()) {
				assert(PoTable.find(id_succ)!=PoTable.end());
				_dMan->getPos()->insert(pair<string,DfgNode*>(PoTable.find(id_succ)->second,pred));
			} else {
				DfgNode* succ = IDTable.find(id_succ)->second;
				switch (succ->getOp()) {
					case DfgOperator::REG: //fall through
					case DfgOperator::EQ: //fall through
					case DfgOperator::LOOPBACK:
						{
							assert(succ->getLeft() ==NULL && succ->getRight() ==NULL);
							assert(pred);
							succ->setKids(pred,pred);
							break;
						}
					case DfgOperator::ADD: //fall through
					case DfgOperator::MUL:
						{
							assert(succ->getRight() ==NULL);
							if (succ->getLeft()) {
								succ->setKids(succ->getLeft(),pred);
							} else {
								assert(succ->getLeft() ==NULL);
								succ->setKids(pred,NULL);
							}
							break;
						}
					case DfgOperator::DIV: //fall through
					case DfgOperator::SUB: //fall through
					case DfgOperator::LSH: //fall through
					case DfgOperator::RSH:
						{
							assert(edgeW);
							assert( (!strcmp(edgeW,"0")) ? (succ->getLeft() ==NULL) : (succ->getRight() ==NULL));
							(!strcmp(edgeW,"0")) ? succ->setKids(pred,succ->getRight()) : succ->setKids(succ->getLeft(),pred);
							break;
						}
					case DfgOperator::CONST: 
						std::cout << "Warning: A constant node is being used as a loopback" << std::endl;
					case DfgOperator::VAR:
					case DfgOperator::VARCONST:
						{
							string loopback_name = Util::strsav(succ->getName());
							// Operand type is being assigned a value by another operand. 
							if (LoopbackTable.find(id_succ)!=LoopbackTable.end()) {
								// Operand comes from an unrolled loop, its loopback
								if (pred->isOnechild()) {
									// the loop was formally sliced by registers, so it is OK to stitch back
									_dMan->getPos()->insert(pair<string,DfgNode*>(loopback_name,pred));
								} else {
									// the loop might have come from other artifact as it has not been properly
									// sliced with registers.
									DfgNode* loopback_node = new DfgNode(_dMan,DfgOperator::LOOPBACK,pred,pred);
									_dMan->registerNode(loopback_node);
									loopback_name += "_loopback";
									_dMan->getPos()->insert(pair<string,DfgNode*>(loopback_name,loopback_node));
								}
							} else {
								// Operand seems to be a temporary node only, we preserve the same DFG format read, 
								// but the temporary operand can be smooth out with the command dfgflatten
								_dMan->getPos()->insert(pair<string,DfgNode*>(loopback_name,pred));
							}
							break;
						}
					default:
						{
							throw(string("09003. Parsed unsupported operator ") + succ->getOpStr());
						}
				}
			}
		}

}
